/*
 * Sentry Video System
 *   controls QoS from clients' requests.
 * 
 *        original by akira-y@is.aist-nara.ac.jp, 1994,1995
 *        modified by k-chinen@is.aist-nara.ac.jp, 1996
 *        modified by satosi-f@is.aist-nara.ac.jp, 1997, 2000
 *
 * $Id: $
 */

/*
 * main.c - main
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <errno.h>

#include "sentry.h"
#include "sentryd.h"

int debug;
int verb;

#ifdef LOGTIME
time_t logtime;
struct tm *log_tm;
#endif

int data_port = SENTRY_DATA_PORT;
int ctrl_port = SENTRY_CTRL_PORT;

int unicast_datasock;
int multicast_datasock;

int maxreadfds;
fd_set readmask;

int maxwritefds;
fd_set writemask;

int require_encode_type = ENCODE_NONE;

char vserver[80] = "localhost";

int 
bind_socket(int sockfd, int port) 
{
  struct sockaddr_in serv_addr;

  bzero((char *)&serv_addr, sizeof(serv_addr));
  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);

  if (bind(sockfd, (struct sockaddr *)&serv_addr, sizeof(serv_addr)) < 0) {
    fprintf(stderr, "sentryd: cannot bind port %d.\n", port);
    close(sockfd);
    return -1;
  }
  return 0;
}

int 
build_listen_port(int port, int type)
{
  int sockfd;
  int flag=1;

  if ((sockfd = socket(AF_INET, type, 0)) < 0) {
    fprintf(stderr, "sentryd: cannot make new TCP socket.\n");
    return -1;
    exit (1);
  }

  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *)&flag, sizeof(flag)) < 0) {
    fprintf(stderr, "sentryd: cannot set sockopt REUSEADDR.\n");
    return -1;
  }

  if (bind_socket(sockfd, port) < 0) {
    fprintf(stderr, "sentryd: please wait several seconds.\n");
    close(sockfd);
    return -1;
    exit (1);
  }

  listen(sockfd, -1);
  if (verb)
    fprintf(stderr, "sentryd: new socket# %d, type# %d, port#, %d.\n", sockfd, type, port);

  return sockfd;
}

int
build_udp_port(int port) 
{
  int sockfd;
  int flag=1;
  int flaglen = sizeof(flag);

  if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    fprintf(stderr, "sentryd: cannot make new UDP socket.\n");
    return -1;
    exit (1);
  }

  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
                 (char *)&flag, sizeof(flag)) < 0) {
    fprintf(stderr, "sentryd: cannot set REUSEADDR\n");
    return -1;
  }


  /*
   * set SNDBUF size:
   *   challenge 16M, 8M, ...., 16K, 8K.
   */
  flag = M(16);
  while(flag >= K(8)) {
    if (setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF,
                 (char *)&flag, flaglen) < 0) {
      fprintf(stderr, "sentryd: fail    set SNDBUF size [%9d bytes].\n", flag);
      flag = flag / 2 ;
    }
    else {
      fprintf(stderr, "sentryd: success set SNDBUF size [%9d bytes].\n", flag);
      break;
    }
  }

  if (verb) {
    if (getsockopt(sockfd, SOL_SOCKET, SO_SNDBUF,
                   (char *)&flag, &flaglen) < 0) {
      fprintf(stderr, "sentryd: cannot get SNDBUF size.\n");
      return -1;
    } else {
      fprintf(stderr, "sentryd: SNDBUF size = %d\n", flag);
    }
  }

  if (bind_socket(sockfd, port) < 0) {
    fprintf(stderr, "sentryd: please wait several seconds.\n");
    close(sockfd);
    return -1;
    exit (1);
  }

  if (verb)
    fprintf(stderr, "sentryd: new udp socket# %d, port# %d.\n", sockfd, port);

  return sockfd;
}

void
reaper()
{
  union wait status;
  while(wait3((int *)&status, WNOHANG, (struct rusage *)0) >= 0)
    ;
}

void
finish_sentryd()
{
  if (verb) {
#ifdef LOGTIME
    logtime = time(NULL);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
	    log_tm->tm_year + 1900, log_tm->tm_mon+1, log_tm->tm_mday,
	    log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stderr, "finish sentryd\nBye.\n");
  }

  syslog(LOG_NOTICE, "sentryd is stopped.\n");
  closelog();
}

void
mainloop(int tsock, int usock)
{
  fd_set readfds;
  int nfds;
  char sysmesg[256];
  int newsockfd;
  struct sockaddr_in from;
  int fromlen;
  int nr;
  char buf[8192];
  extern void init_req_val(), inti_client();

  (void)init_req_val();
  (void)init_client();

  FD_ZERO(&readmask);
  FD_SET(tsock, &readmask);
  FD_SET(usock, &readmask);
  maxreadfds = MAX(tsock, usock);

  FD_ZERO(&writemask);
  FD_SET(unicast_datasock, &writemask);
  FD_SET(multicast_datasock, &writemask);
  maxwritefds = MAX(unicast_datasock, multicast_datasock);

  while (1) {
    readfds = readmask;

    nfds = select(maxreadfds+1, &readfds, (fd_set *)NULL, (fd_set *)NULL, (struct timeval *)NULL);
    if (nfds <= 0) {
      sprintf(sysmesg, "sentryd: select: %s", sys_errlist[errno]);
      syslog(LOG_ERR, sysmesg);
      continue;
    }

    if (FD_ISSET(tsock, &readmask)) {
      FD_CLR(tsock, &readmask);
      newsockfd = accept(tsock, (struct sockaddr *)&from, &fromlen);
      if (newsockfd < 0)
	goto udp_check;

      if (verb) 
	fprintf(stderr, "TCP connection is arrved.\n");


      if (add_client(newsockfd, &from) < 0) {
	fprintf (stderr, "close connection port# %d.\n", newsockfd);
	shutdown(newsockfd, 2);
	close(newsockfd);
      } else {
	FD_SET(newsockfd, &readmask);
	maxreadfds = MAX(maxreadfds, newsockfd);

	printf("sentryd: tcp sock# %d is connected from %s:%d.\n", newsockfd, inet_ntoa(from.sin_addr), ntohs(from.sin_port));
      }
    }

  udp_check:
    if (FD_ISSET(usock, &readmask)) {
      if (verb)
	fprintf (stderr, "UDP is arrive.\n");

      get_udp_request(usock);

      printf("sentryd: QoS request are arrived to udp port# %d from %s:%d\n", usock, inet_ntoa(from.sin_addr), ntohs(from.sin_port));


    }else {
      /*
       * QoS requests are arrived from TCP port.
       *   not implemented yet... :)
       */
    }

  } /* end-of-while */
  /* NOT REACHER */
}


int
main(int argc, char **argv)
{
  int tcpsock, udpsock;
  int c;
  extern char *optarg;
  extern void usage(), finish_sentryd();

#ifdef DEBUG
  debug = 1;
#endif
#ifdef VERBOSE
  verb = 1;
#endif /* VERBOSE */

  if (verb) 
    fprintf(stdout, "sentryd start");

  openlog("sentryd", LOG_PID, LOG_DAEMON);
  syslog(LOG_NOTICE, "sentryd start");
  fprintf(stderr, "sentryd start\n");

  if (verb) {
#ifdef LOGTIME
    logtime = time(null);
    log_tm = localtime(&logtime);
    fprintf(stderr, "[%4d/%02d/%02d %02d:%02d:%02d] ",
	    log_tm->tm_year+1900, log_tm->tm_mon+1, log_tm->tm_mday,
	    log_tm->tm_hour, log_tm->tm_min, log_tm->tm_sec);
#endif
    fprintf(stderr, "sentryd start\n");
  }

  signal(SIGQUIT, finish_sentryd);
  signal(SIGTERM, finish_sentryd);
  signal(SIGINT, finish_sentryd);

  require_encode_type = ENCODE_NONE;
  
  while ((c = getopt(argc, argv, "hDvjs:")) != -1) {
    switch(c) {
    case 'h':
      usage();
      exit (1); /* not reached */
    case 'D':
      debug ++;
      break;
    case 'v':
      verb = 1;
      break;
    case 'j':
      require_encode_type = ENCODE_JPEG;
      break;
    case 'p':
      data_port = atoi(optarg);
      ctrl_port = data_port + 1;
      break;
    }
  }
  
  /*
   * video setup
   */
  if (check_video(vserver)) {
    fprintf(stderr, "sentryd: sorry, video(%s) is not avaliable.\n", "localhost");
    exit (2);
  }

  tcpsock = build_listen_port(ctrl_port, SOCK_STREAM);
  udpsock = build_listen_port(ctrl_port, SOCK_DGRAM);

  unicast_datasock = build_udp_port(0);
  multicast_datasock = build_udp_port(0);

  if (tcpsock < 0 || udpsock < 0)
    exit (1);
  
  (void)mainloop(tcpsock, udpsock);
  exit (0);
}

void
usage() {
  printf("usage: sentryd [options]\n\n");

  printf("option:\n");
  printf("       -h       help\n");
  printf("       -j       enable jpeg encode(deafult=disable)\n");
  printf("       -p port  data port(default=%d)\n", SENTRY_DATA_PORT);
  printf("       -D       debug mode\n");
  printf("       -v       verbose mode\n");
  printf("\n");

  exit (1);
}



  
  
